{
import java.util.*;
import java.io.*;
}

/** Read in an ANTLR grammar and build an AST.  For now, I'm ignoring
 *  anything except the rules.
 *
 *  Terence Parr
 *  University of San Francisco
 *  August 19, 2003
 */
class ANTLRParser extends Parser;
options {
    buildAST = true;
	exportVocab=ANTLR;
    ASTLabelType="GrammarAST";
	k=2;
}

tokens {
	"tokens";
    LEXER;
    PARSER;
    TREE_PARSER;
    RULE;
    BLOCK;
    OPTIONAL;
    CLOSURE;
    POSITIVE_CLOSURE;
    SYNPRED;
    TOKEN_RANGE;
    CHAR_RANGE;
    NOT;
    EPSILON;
    ALT;
    EOR;
}

{
// a bunch of (non thread safe and terrible OO design) maps/lists
// that you might find useful.  I'll refactor later...this stuff
// is in my "playground" not my "repository" yet. :)  TJP 8-19-2003
public static Hashtable ruleToTreeMap = new Hashtable();
public static List subruleToTreeList = new ArrayList();
public int tokenType = Edge.MIN_TOKEN_TYPE;
public static Hashtable tokenToTypeMap = new Hashtable();
public static Vector typeToTokenList = new Vector();

private int ruleIndex = 1;
public static Hashtable ruleToIndexMap = new Hashtable();
public static Vector ruleIndexToRuleList = new Vector();

public void defineToken(String text) {
		tokenToTypeMap.put(text, new Integer(tokenType));
		typeToTokenList.setSize(tokenType+1);
		typeToTokenList.set(tokenType, text);
		tokenType++;
}

public void defineCharLiteralToken(String charLiteral) {
		int c = charLiteral.charAt(1);
		tokenToTypeMap.put(charLiteral, new Integer(c));
}

public void defineRule(String ruleName) {
		ruleToIndexMap.put(ruleName, new Integer(ruleIndex));
		ruleIndexToRuleList.setSize(ruleIndex+1);
		ruleIndexToRuleList.set(ruleIndex, ruleName);
		ruleIndex++;
}

public void init() {
	typeToTokenList.setSize(Edge.EOF+1);
    typeToTokenList.set(0, "<INVALID>");
    typeToTokenList.set(Edge.EOF, "<EOF>");
    tokenToTypeMap.put("<INVALID>", new Integer(0));
    tokenToTypeMap.put("<EOF>", new Integer(Edge.EOF));
}

public void finish() {
}

}

grammar!
	{GrammarAST c=null;}
   :{init();}
	( 	"header"
	 	(n:STRING_LITERAL)?
	 	h:ACTION
	)*
	( fileOptionsSpec )?
    ( cd:classDef {if (c==null) {c=#cd;} else {c.setNextSibling(#cd);}} )*
	EOF
    {#grammar = c;}
    {finish();}
	;

classDef!
	:
	( a:ACTION )?
	( d:DOC_COMMENT  )?
	(	("lexclass" | "class" id "extends" "Lexer" ) => lexer:lexerSpec
        {#classDef = #[LEXER,"lexer"];}
	|	( "class" id "extends" "TreeParser" ) => treeparser:treeParserSpec
        {#classDef = #[TREE_PARSER,"tree_parser"];}
	|	parser:parserSpec
        {#classDef = #[PARSER,"parser"];}
	)
	r:rules
    {#classDef = #(#classDef, r);}
	;

fileOptionsSpec

	:	OPTIONS
		(
			 id
			ASSIGN
			 optionValue
			
			SEMI
		)*
		RCURLY
	;

parserOptionsSpec

	:	OPTIONS
		(
			 id
			ASSIGN
			 optionValue
			
			SEMI
		)*
		RCURLY
	;

treeParserOptionsSpec

	:	OPTIONS
		(
			 id
			ASSIGN
			 optionValue
			
			SEMI
		)*
		RCURLY
	;

lexerOptionsSpec

	:
	OPTIONS
	(	// Special case for vocabulary option because it has a bit-set
		"charVocabulary"
		ASSIGN
		 charSet
		SEMI
		

	|	 id
		ASSIGN
		 optionValue
		
		SEMI
	)*
	RCURLY
	;

subruleOptionsSpec

	:	OPTIONS
		(	 id
			ASSIGN
			 optionValue
			
			SEMI
		)*
		RCURLY
	;

// optionValue returns a Token which may be one of several things:
//    STRING_LITERAL -- a quoted string
//    CHAR_LITERAL -- a single quoted character
//		INT -- an integer
//		RULE_REF or TOKEN_REF -- an identifier
optionValue
	:	 qualifiedID
	|	sl:STRING_LITERAL 
	|	cl:CHAR_LITERAL 
	|	il:INT 
	;

charSet
	:
		 setBlockElement
		(
			OR
			 setBlockElement
			
		)*
	;

setBlockElement
	:
	c1:CHAR_LITERAL
	(
		RANGE c2:CHAR_LITERAL
	)?
	;

tokensSpec
	:	TOKENS
			(	(	
					t1:TOKEN_REF
					( ASSIGN s1:STRING_LITERAL )?
					
					(tokensSpecOptions)?
				|	s3:STRING_LITERAL
					
					(tokensSpecOptions)?
				)
				SEMI
			)+
		RCURLY
	;

tokensSpecOptions
	:	OPEN_ELEMENT_OPTION
		id ASSIGN optionValue
		
		(
			SEMI
			id ASSIGN optionValue
			
		)*
		CLOSE_ELEMENT_OPTION
	;

superClass
	:	LPAREN
			
			(TOKEN_REF|RULE_REF)
		RPAREN
	;

parserSpec
	:	"class"^
		 id
		(	"extends" "Parser"
			(superClass)?
		|	
		)
		
		SEMI!
		(parserOptionsSpec!)?
		
		(tokensSpec!)?
		( ACTION! )?
	;

lexerSpec
	:   (	lc:"lexclass"
			 id
		|	"class"^
			 id
			"extends"
			"Lexer"
			(superClass)?
		)
		
		SEMI!
		(lexerOptionsSpec!)?
		
		(tokensSpec!)?
		( ACTION! )?
	;

treeParserSpec
	:	"class"^
		 id
		"extends"
		"TreeParser"
		(superClass)?
		
		SEMI!
		(treeParserOptionsSpec!)?
		
		(tokensSpec!)?
		( ACTION! )?
	;

rules
    :   (
			options {
				// limitation of appox LL(k) says ambig upon
				// DOC_COMMENT TOKEN_REF, but that's an impossible sequence
				warnWhenFollowAmbig=false;
			}
		:	rule
		)+
    ;

rule!
	:
	(	d:DOC_COMMENT	
	)?
	(	p1:"protected"	
	|	p2:"public"		
	|	p3:"private"	
	)?
	ruleName:id    {defineRule(#ruleName.getText());}
	( BANG  )?
	( aa:ARG_ACTION   )?
	( "returns" rt:ARG_ACTION  )?
	( throwsSpec )?
	( ruleOptionsSpec )?
	(a:ACTION )?
	COLON b:block SEMI
	( exceptionGroup )?
    {
    GrammarAST eor = #[EOR,"<end-of-rule>"];
    eor.setEnclosingRule(#ruleName.getText());
    #rule = #(#[RULE,"rule"],#ruleName,#b,eor);
    }
    {ruleToTreeMap.put(#ruleName.getText(), #rule);}
    {subruleToTreeList.add(#b);} // add block to subrule list also
	;

ruleOptionsSpec
	:	OPTIONS
		(
			 id
			ASSIGN
			 optionValue
			
			SEMI
		)*
		RCURLY
	;

throwsSpec
	:	"throws" id 
		( COMMA id  )*
		
	;

// Build #(BLOCK ( #(ALT .) )+ )

block
{
GrammarAST root = #[BLOCK];
}
    :   a1:alternative
        ( OR! a2:alternative
        )*
        {#block = #(root,#block);}
    ;

alternative // [GrammarAST enclosingBlock]
    :   (BANG!)? ( element )+ ( exceptionSpecNoLabel! )?
        {
            if ( #alternative==null ) {
                #alternative = #(#[ALT,"ALT"],#[EPSILON,"epsilon"]);
            }
            else {
                #alternative = #(#[ALT,"ALT"], #alternative);
            }
        }
    |   {#alternative = #(#[ALT,"ALT"],#[EPSILON,"epsilon"]);}
    ;

exceptionGroup
	:	( exceptionSpec )+
    ;

exceptionSpec
    :   "exception" ( ARG_ACTION  )?
        ( exceptionHandler )*
    ;

exceptionSpecNoLabel
    :   "exception" ( exceptionHandler )*
    ;

exceptionHandler
   :
   "catch"
   ARG_ACTION
   ACTION
   ;

element
	:	elementNoOptionSpec (elementOptionSpec!)?
	;

elementOptionSpec
	:	OPEN_ELEMENT_OPTION
		id ASSIGN optionValue
		(
			SEMI
			id ASSIGN optionValue
			
		)*
		CLOSE_ELEMENT_OPTION
	;

elementNoOptionSpec!
	:	id
		ASSIGN
		( id COLON  )?
		(	rr:RULE_REF {#elementNoOptionSpec = #rr;}
			( ARG_ACTION  )?
			( BANG  )?
		|	// this syntax only valid for lexer
			tr:TOKEN_REF {#elementNoOptionSpec = #tr;}
			( ARG_ACTION  )?
		)
	|
		(id COLON  )?
		(	r2:RULE_REF {#elementNoOptionSpec = #r2;}
			( ARG_ACTION  )?
			( BANG  )?
		|   r:range {#elementNoOptionSpec = #r;}
		|   t:terminal  {#elementNoOptionSpec = #t;}
		|	NOT_OP
			(	nt:notTerminal  {#elementNoOptionSpec = #(#[NOT,"~"],#nt);}
			|	e1:ebnf         {#elementNoOptionSpec = #(#[NOT,"~"],#e1);}
			)
		|	e2:ebnf             {#elementNoOptionSpec = #e2;}
		)
	|
		ACTION//	            {#elementNoOptionSpec = #a;}
	|
		SEMPRED//	            {#elementNoOptionSpec = #p;}
	|
		t3:tree                 {#elementNoOptionSpec = #t3;}
	;

tree :
	TREE_BEGIN^
        rootNode
        ( element )+
    RPAREN!
	;

rootNode
	:   (id! COLON!)?
		terminal
	;

ebnf!
	:	lp:LPAREN
		(
			// 2nd alt and optional branch ambig due to
			// linear approx LL(2) issue.  COLON ACTION
			// matched correctly in 2nd alt.
			options {
				warnWhenFollowAmbig = false;
			}
		:
			subruleOptionsSpec ( ACTION )? COLON
		|	ACTION COLON
		)?

		b:block
        RPAREN
		(	(	QUESTION    {#b.setType(OPTIONAL); #ebnf=#b;}
			|	STAR	    {#b.setType(CLOSURE); #ebnf=#b;}
			|	PLUS	    {#b.setType(POSITIVE_CLOSURE); #ebnf=#b;}
			)
			( BANG )?
		|   IMPLIES	        {#b.setType(SYNPRED); #ebnf=#b;}
        |                   {#ebnf = #b;}
		)
        {subruleToTreeList.add(#ebnf);}
	;

ast_type_spec
	:	(	CARET
		|	BANG 
		)?
	;

range!
	:	crLeft:CHAR_LITERAL RANGE crRight:CHAR_LITERAL ( BANG )?
		{#range = #(#[CHAR_RANGE,".."], #crLeft, #crRight);}
	|
		(t:TOKEN_REF | u:STRING_LITERAL )
		RANGE
		(v:TOKEN_REF | w:STRING_LITERAL )
		ast_type_spec
		{
        AST a = (#t!=null) ? #t : #u;
        AST b = (#v!=null) ? #v : #w;
        #range = #(#[TOKEN_RANGE,".."], a, b);
        }
	;

terminal
	:
		cl:CHAR_LITERAL ( BANG! )? // {defineCharLiteralToken(cl.getText());}
	|
		tr:TOKEN_REF
		{
		defineToken(tr.getText());
		}
		ast_type_spec!
		// Args are only valid for lexer
		( ARG_ACTION! )?
	|
		sl:STRING_LITERAL
		{
		defineToken(sl.getText());
		}
		ast_type_spec!
	|
		wi:WILDCARD
		 ast_type_spec!
	;

notTerminal
	:   cl:CHAR_LITERAL
		( BANG!  )?
	|
		tr:TOKEN_REF
		 ast_type_spec!
	;

/** Match a.b.c.d qualified ids; WILDCARD here is overloaded as
 *  id separator; that is, I need a reference to the '.' token.
 */
qualifiedID
	:	id 
		(	WILDCARD id
			
		)*
	;

id  :	TOKEN_REF			
	|	RULE_REF			
	;

class ANTLRLexer extends Lexer;
options {
	k=2;
	exportVocab=ANTLR;
	testLiterals=false;
	interactive=true;
	charVocabulary='\003'..'\377';
}

tokens {
	"options";
}

{
	/**Convert 'c' to an integer char value. */
	public static int escapeCharValue(String cs) {
		//System.out.println("escapeCharValue("+cs+")");
		if ( cs.charAt(1)!='\\' ) return 0;
		switch ( cs.charAt(2) ) {
		case 'b' : return '\b';
		case 'r' : return '\r';
		case 't' : return '\t';
		case 'n' : return '\n';
		case 'f' : return '\f';
		case '"' : return '\"';
		case '\'' :return '\'';
		case '\\' :return '\\';

		case 'u' :
			// Unicode char
			if (cs.length() != 8) {
				return 0;
			}
			else {
				return
					Character.digit(cs.charAt(3), 16) * 16 * 16 * 16 +
					Character.digit(cs.charAt(4), 16) * 16 * 16 +
					Character.digit(cs.charAt(5), 16) * 16 +
					Character.digit(cs.charAt(6), 16);
			}

		case '0' :
		case '1' :
		case '2' :
		case '3' :
			if ( cs.length()>5 && Character.isDigit(cs.charAt(4)) ) {
				return (cs.charAt(2)-'0')*8*8 + (cs.charAt(3)-'0')*8 + (cs.charAt(4)-'0');
			}
			if ( cs.length()>4 && Character.isDigit(cs.charAt(3)) ) {
				return (cs.charAt(2)-'0')*8 + (cs.charAt(3)-'0');
			}
			return cs.charAt(2)-'0';

		case '4' :
		case '5' :
		case '6' :
		case '7' :
			if ( cs.length()>4 && Character.isDigit(cs.charAt(3)) ) {
				return (cs.charAt(2)-'0')*8 + (cs.charAt(3)-'0');
			}
			return cs.charAt(2)-'0';

		default :
			return 0;
		}
	}

	public static int tokenTypeForCharLiteral(String lit) {
		if ( lit.length()>3 ) {  // does char contain escape?
			return escapeCharValue(lit);
		}
		else {
			return lit.charAt(1);
		}
	}
}

WS	:	(	/*	'\r' '\n' can be matched in one alternative or by matching
				'\r' in one iteration and '\n' in another.  I am trying to
				handle any flavor of newline that comes in, but the language
				that allows both "\r\n" and "\r" and "\n" to all be valid
				newline is ambiguous.  Consequently, the resulting grammar
				must be ambiguous.  I'm shutting this warning off.
			 */
			options {
				generateAmbigWarnings=false;
			}
		:	' '
		|	'\t'
		|	'\r' '\n'	{newline();}
		|	'\r'		{newline();}
		|	'\n'		{newline();}
		)
		{ $setType(Token.SKIP); }
	;

COMMENT :
	( SL_COMMENT | t:ML_COMMENT {$setType(t.getType());} )
	{if ( _ttype != DOC_COMMENT ) $setType(Token.SKIP);}
	;

protected
SL_COMMENT :
	"//"
	( ~('\n'|'\r') )*
	(
		/*	'\r' '\n' can be matched in one alternative or by matching
			'\r' and then in the next token.  The language
			that allows both "\r\n" and "\r" and "\n" to all be valid
			newline is ambiguous.  Consequently, the resulting grammar
			must be ambiguous.  I'm shutting this warning off.
		 */
			options {
				generateAmbigWarnings=false;
			}
		:	'\r' '\n'
		|	'\r'
		|	'\n'
	)
	{ newline(); }
	;

protected
ML_COMMENT :
	"/*"
	(	{ LA(2)!='/' }? '*' {$setType(DOC_COMMENT);}
	|
	)
	(
		/*	'\r' '\n' can be matched in one alternative or by matching
			'\r' and then in the next token.  The language
			that allows both "\r\n" and "\r" and "\n" to all be valid
			newline is ambiguous.  Consequently, the resulting grammar
			must be ambiguous.  I'm shutting this warning off.
		 */
		options {
			greedy=false;  // make it exit upon "*/"
			generateAmbigWarnings=false; // shut off newline errors
		}
	:	'\r' '\n'	{newline();}
	|	'\r'		{newline();}
	|	'\n'		{newline();}
	|	~('\n'|'\r')
	)*
	"*/"
	;

OPEN_ELEMENT_OPTION
	:	'<'
	;

CLOSE_ELEMENT_OPTION
	:	'>'
	;

COMMA : ',';

QUESTION :	'?' ;

TREE_BEGIN : "#(" ;

LPAREN:	'(' ;

RPAREN:	')' ;

COLON :	':' ;

STAR:	'*' ;

PLUS:	'+' ;

ASSIGN : '=' ;

IMPLIES : "=>" ;

SEMI:	';' ;

CARET : '^' ;

BANG : '!' ;

OR	:	'|' ;

WILDCARD : '.' ;

RANGE : ".." ;

NOT_OP :	'~' ;

RCURLY:	'}'	;

CHAR_LITERAL
	:	'\'' (ESC|~'\'') '\''
	;

STRING_LITERAL
	:	'"' (ESC|~'"')* '"'
	;

protected
ESC	:	'\\'
		(	'n'
		|	'r'
		|	't'
		|	'b'
		|	'f'
		|	'w'
		|	'a'
		|	'"'
		|	'\''
		|	'\\'
		|	('0'..'3')
			(
				options {
					warnWhenFollowAmbig = false;
				}
			:
	('0'..'9')
				(
					options {
						warnWhenFollowAmbig = false;
					}
				:
	'0'..'9'
				)?
			)?
		|	('4'..'7')
			(
				options {
					warnWhenFollowAmbig = false;
				}
			:
	('0'..'9')
			)?
		|	'u' XDIGIT XDIGIT XDIGIT XDIGIT
		)
	;

protected
DIGIT
	:	'0'..'9'
	;

protected
XDIGIT :
		'0' .. '9'
	|	'a' .. 'f'
	|	'A' .. 'F'
	;

INT	:	('0'..'9')+
	;

ARG_ACTION
   :
	NESTED_ARG_ACTION
	;

protected
NESTED_ARG_ACTION :
	'['
	(
		/*	'\r' '\n' can be matched in one alternative or by matching
			'\r' and then '\n' in the next iteration.
		 */
		options {
			generateAmbigWarnings=false; // shut off newline errors
		}
	:	NESTED_ARG_ACTION
	|	'\r' '\n'	{newline();}
	|	'\r'		{newline();}
	|	'\n'		{newline();}
	|	CHAR_LITERAL
	|	STRING_LITERAL
	|	~']'
	)*
	']'
	;

ACTION
{int actionLine=getLine(); int actionColumn = getColumn(); }
	:	NESTED_ACTION
		(	'?'	{_ttype = SEMPRED;} )?
		{
			CommonToken t = new CommonToken(_ttype,$getText);
			t.setLine(actionLine);			// set action line to start
			t.setColumn(actionColumn);
			$setToken(t);
		}
	;

protected
NESTED_ACTION :
	'{'
	(
		options {
			greedy = false; // exit upon '}'
		}
	:
		(
			options {
				generateAmbigWarnings = false; // shut off newline warning
			}
		:	'\r' '\n'	{newline();}
		|	'\r' 		{newline();}
		|	'\n'		{newline();}
		)
	|	NESTED_ACTION
	|	CHAR_LITERAL
	|	COMMENT
	|	STRING_LITERAL
	|	.
	)*
	'}'
   ;

TOKEN_REF
options { testLiterals = true; }
	:	'A'..'Z'
		(	// scarf as many letters/numbers as you can
			options {
				warnWhenFollowAmbig=false;
			}
		:
			'a'..'z'|'A'..'Z'|'_'|'0'..'9'
		)*
	;

// we get a warning here when looking for options '{', but it works right
RULE_REF
{
	int t=0;
}
	:	t=INTERNAL_RULE_REF {_ttype=t;}
		(	{t==LITERAL_options}? WS_LOOP ('{' {_ttype = OPTIONS;})?
		|	{t==LITERAL_tokens}? WS_LOOP ('{' {_ttype = TOKENS;})?
		|
		)
	;

protected
WS_LOOP
	:	(	// grab as much WS as you can
			options {
				greedy=true;
			}
		:
			WS
		|	COMMENT
		)*
	;

protected
INTERNAL_RULE_REF returns [int t]
{
	t = RULE_REF;
}
	:	'a'..'z'
		(	// scarf as many letters/numbers as you can
			options {
				warnWhenFollowAmbig=false;
			}
		:
			'a'..'z'|'A'..'Z'|'_'|'0'..'9'
		)*
		{t = testLiteralsTable(t);}
	;

protected
WS_OPT :
	(WS)?
	;
