header {
	import java.util.*;
}

/** Print out a grammar (no pretty printing).
 *
 *  Terence Parr
 *  University of San Francisco
 *  August 19, 2003
 */
class ANTLRTreePrinter extends TreeParser;

options {
	importVocab = ANTLR;
    codeGenBitsetTestThreshold=999;
}

{
    StringBuffer buf = new StringBuffer(300);

    public void out(String s) {
        buf.append(s);
    }

    /** Parser error-reporting function can be overridden in subclass */
    public void reportError(RecognitionException ex) {
        System.out.println("print: "+ex.toString());
    }

    /** Parser error-reporting function can be overridden in subclass */
    public void reportError(String s) {
        System.out.println("print: error: " + s);
    }

	/** Normalize a grammar print out by removing all double spaces
	 *  and trailing/beginning stuff.  FOr example, convert
	 *
	 *  ( A  |  B  |  C )*
	 *
	 *  to
	 *
	 *  ( A | B | C )*
	 */
	public static String normalize(String g) {
	    StringTokenizer st = new StringTokenizer(g, " ", false);
		StringBuffer buf = new StringBuffer();
		while ( st.hasMoreTokens() ) {
			String w = st.nextToken();
			buf.append(w);
			buf.append(" ");
		}
		return buf.toString().trim();
	}
}

/** Call this to figure out how to print */
toString returns [String s=null]
    :   (   classDef
        |   rule
        |   alternative
        |   element
        |   EOR {s="EOR";}
        )
        {return normalize(buf.toString());}
    ;

// --------------

grammar
    :   ( classDef ) // ignore 2nd (lexer) grammar for now
    ;

classDef
    :   #( PARSER rules )
    |   #( TREE_PARSER rules )
    |   #( LEXER rules )
    ;

rules
    :   ( rule )+
    ;

rule
    :   #( RULE r:RULE_REF {out(r+" : ");} #(BLOCK block) EOR {out(";\n");} )
    ;

block
    :   {out(" (");}
        alternative ( {out(" | ");} alternative)*
        {out(")");}
    ;

alternative
    :   #( ALT (element)+ )
    ;

element
    :   atom
    |   #(NOT {out("~");} element) 
    |   #(RANGE atom {out("..");} atom)
    |   #(CHAR_RANGE CHAR_LITERAL {out("..");} CHAR_LITERAL)
    |   ebnf
    |   tree
    |   #( SYNPRED block ) {out("=>");}
    |   ACTION
    |   SEMPRED
    |   EPSILON {out(" epsilon ");}
    ;

ebnf:   #( BLOCK block ) {out(" ");}
    |   #( OPTIONAL block ) {out("? ");}
    |   #( CLOSURE block )  {out("* ");}
    |   #( POSITIVE_CLOSURE block ) {out("+ ");}
    ;

tree:   #(TREE_BEGIN {out(" #(");} atom (element)* {out(") ");} )
    ;

atom
{out(" "+#atom.toString()+" ");}
    :   RULE_REF
    |   TOKEN_REF
    |   CHAR_LITERAL
    |   STRING_LITERAL
    |   WILD_CARD
    ;

